from .scatter import scatter
from .velocity import velocity
from .velocity_embedding import velocity_embedding
from .velocity_embedding_grid import velocity_embedding_grid
from .velocity_embedding_stream import velocity_embedding_stream
from .velocity_graph import velocity_graph
from .heatmap import heatmap
from .utils import hist, plot
from .simulation import simulation
from scanpy.api.pl import paga, paga_compare, rank_genes_groups
